package Algorithms.Animated.BK;

import java.util.ArrayList;

import Algorithms.Animated.AlgorithmStage;
import Algorithms.Animated.BackwardAction;
import Algorithms.Animated.BK.ExtremalLayoutCalc.LayoutType;
import Model.LayeredGraphNode;

/**
 * The stage of compacting the layout.
 * @author kolja
 *
 */
public class Compaction implements AlgorithmStage{

	private enum CompactionState
	{
		PLACE_BLOCKS,
		APPLY_SHIFT
	}
	
	private class StackFrame
	{
		public LayeredGraphNode v;
		public LayeredGraphNode u;
		public LayeredGraphNode w;
	}
	
	private CompactionState state;
	private LayeredGraphNode graph;
	private int vIndex;
	
	private ArrayList< StackFrame > stack;
	private ArrayList< BackwardAction > actions;
	private LayoutType layout;
	
	public Compaction( LayeredGraphNode graph, LayoutType layout )
	{
	    this.layout = layout;
		this.graph = graph;
		state = CompactionState.PLACE_BLOCKS;
		stack = new ArrayList<>();
		vIndex = 0;
		actions = new ArrayList<>();
	}
	
	public double calcSpacing()
	{
		double max = 0;
		for( LayeredGraphNode n : graph.getContainedNodes() )
			max = Math.max( max, n.getWidth( layout ) );
		return max + 50;
	}
	
	@Override
	public StageStatus forwardStep() {
		int acSize = actions.size();
		if( state == CompactionState.PLACE_BLOCKS )
		{
			if( stack.size() == 0 )
			{
				ArrayList< LayeredGraphNode > nodes = graph.getContainedNodes();
				boolean found = false;
				int oldVIndex = vIndex;
				for( ; vIndex < nodes.size(); vIndex++ )
				{
					if( nodes.get( vIndex ).isXUndefined( layout ) && nodes.get( vIndex ) == nodes.get( vIndex ).getRoot( layout ) )
					{
						found = true;
						break;
					}
				}
				if( !found )
				{
					state = CompactionState.APPLY_SHIFT;
					vIndex = 0;
					actions.add( 0, ()-> {
						vIndex = oldVIndex;
						state = CompactionState.PLACE_BLOCKS;
					} );
				}
				else
				{
					StackFrame f = new StackFrame();
					f.v = nodes.get( vIndex );
					double oldX = f.v.getX( layout );
					f.v.setX( 0, true, layout );
					f.v.setSelected( layout );
					f.w = f.v;
					stack.add( 0, f );
					actions.add( 0, ()-> {
						stack.get( 0 ).v.setX( oldX, false, layout );
						stack.get( 0 ).v.setSelected( layout );
						stack.remove( 0 );
						vIndex = oldVIndex;
						state = CompactionState.PLACE_BLOCKS;
					});
				}
			}
			else
			{
				StackFrame sf = stack.get( 0 );
				if( sf.u == null )
				{
					if( graph.getContainedLayers().get( sf.w.getLayer() ).indexOf( sf.w ) >= 1 )
					{
						sf.u = graph.getContainedLayers().get( sf.w.getLayer() ).get( graph.getContainedLayers().get( sf.w.getLayer() ).indexOf( sf.w ) - 1 ).getRoot( layout );
						if( sf.u.isXUndefined( layout ) )
						{
							StackFrame nsf = new StackFrame();
							nsf.v = sf.u;
							double oldX = nsf.v.getX( layout );
							nsf.v.setX( 0, true, layout );
							nsf.v.setSelected( layout );
							nsf.w = nsf.v;
							stack.add( 0, nsf );
							actions.add( 0, ()-> {
								stack.get( 0 ).v.setX( oldX, false, layout );
								stack.get( 0 ).v.setSelected( layout );
								stack.remove( 0 );
								stack.get( 0 ).u = null;
							});
						}
						else
						{
							sf.u.setSelected( layout );
							actions.add( 0, ()-> {
								stack.get( 0 ).u = null;
							});
						}
					}
					else
					{
						LayeredGraphNode oldW = sf.w;
						sf.v.setSelected( layout );
						sf.w = sf.w.getAlignedTo( layout );
						if( sf.w == sf.v )
						{
							stack.remove( 0 );
							actions.add( 0, ()-> {
								stack.add( 0, sf );
								sf.v.setSelected( layout );
								sf.w = oldW;
							});
						}
						else
						{
							actions.add( 0, ()-> {
								sf.w = oldW;
								sf.v.setSelected( layout );
							});
						}
					}	
				}
				else
				{
					LayeredGraphNode oldSink = sf.v.getSink( layout );
					if( sf.v.getSink( layout ) == sf.v )
						sf.v.setSink( sf.u.getSink( layout ), layout );
					LayeredGraphNode sinkOfU = sf.u.getSink( layout );
					double oldShift = sinkOfU.getShift( layout );
					double oldX = sf.v.getX( layout );
					boolean oldDef = !sf.v.isXUndefined( layout );
					sf.v.setSelected( layout );
					if( sf.v.getSink( layout ) != sf.u.getSink( layout ) )
						sf.u.getSink( layout ).setShift( Math.min( sf.u.getSink( layout ).getShift( layout ), sf.v.getX( layout ) - sf.u.getX( layout ) - calcSpacing() ), layout );
					else
						sf.v.setX( Math.max( sf.v.getX( layout ), sf.u.getX( layout ) + calcSpacing() ), true, layout );
					LayeredGraphNode oldW = sf.w;
					sf.w = sf.w.getAlignedTo( layout );
					LayeredGraphNode oldU = sf.u;
					sf.u = null;
					if( sf.w == sf.v )
					{
						stack.remove( 0 );
						actions.add( 0, ()-> {
							stack.add( 0, sf );
							stack.get( 0 ).v.setSink(  oldSink, layout );
							sinkOfU.setShift( oldShift, layout );
							sf.u = oldU;
							sf.v.setX( oldX, oldDef, layout );
							sf.v.setSelected( layout );
							sf.w = oldW;
						});
					}
					else
					{
						actions.add( 0, ()-> {
							stack.get( 0 ).v.setSink(  oldSink, layout );
							sinkOfU.setShift( oldShift, layout );
							sf.u = oldU;
							sf.v.setX( oldX, oldDef, layout );
							sf.v.setSelected( layout );
							sf.w = oldW;
						});
					}
				}
			}
		}
		else if( state == CompactionState.APPLY_SHIFT )
		{
			LayeredGraphNode v = graph.getContainedNodes().get( vIndex );
			double oldX = v.getX( layout );
			boolean oldDef = !v.isXUndefined( layout );
			v.setSelected( layout );
			v.setX( v.getRoot( layout ).getX( layout ), true, layout );
			if( v == v.getRoot( layout ) && v.getSink( layout ).getShift( layout ) < Double.POSITIVE_INFINITY )
				v.setX( v.getX( layout ) + v.getSink( layout ).getShift( layout ), true, layout );
			actions.add( 0, ()-> {
				v.setX( oldX, oldDef, layout );
				v.setSelected( layout );
				vIndex--;
			} );
			vIndex++;
			if( vIndex >= graph.getContainedNodes().size() )
				return StageStatus.FINISHED;
		}
		if( actions.size() != acSize + 1 )
			System.out.println( "ERROR" );
		return StageStatus.UNFINISHED;
	}

	@Override
	public StageStatus backwardStep() {
		if( actions.size() == 0 )
			return StageStatus.FINISHED;
		actions.get( 0 ).reverse();
		actions.remove( 0 );
		return StageStatus.UNFINISHED;
	}

}
